/*
 * sdr - 'SCSI Direct Read'
 *   Copyright (C)1999 Tachibana Eriko
 */

#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/dos.h>
#include <sys/scsi.h>
#include <sys/xglob.h>

#include "sdr.h"


/* File scope function */
static int
get_top_blk (int unit_no, int sec_size, int scsi_id, int size)
{
  unsigned char buf[256 << size];

  /* 擪ZN^ǂݍ */
  if (_scsi_readext (0, 1, scsi_id, size, buf) < 0)
    return -1;

  /* X680x0 */
  if (memcmp ("X68SCSI1", buf, 8) == 0)
    {
      int i;

      /* ItZbg 2048 oCg̃ZN^ǂݍ */
      if (_scsi_readext (2048 >> 8 >> size, 1, scsi_id, size, buf) < 0)
	return -1;
      if (memcmp ("X68K", buf, 4) != 0)
	return -1;

      for (i = 1; i < 16; ++i)
	{
	  if (memcmp ("Human68k", &buf[i * 16 + 0], 8) != 0)
	    continue;			/*  Human68k p[eBV */
	  if (buf[i * 16 + 8] & 1)
	    continue;			/* ANeBuȃp[eBV */

	  if (--unit_no < 0)
	    {
	      int r = (*(int*) &buf[i * 16 + 8]) & 0x00ffffff;

	      return (r << (3 - size)) >> 1;
	    }
	}

      return -1;
    }

#if 0					/* Ή */
  /* IBM-FDISK */
  else if ((buf[0] == 0xb8 || buf[0] == 0xfa || buf[0] == 0x00)
	&& buf[sec_size - 2] == 0x55 && buf[sec_size - 1] == 0xaa)
    {
      return -1;
    }
#endif

  /* IBM-PC Super-Floppy */
  else if (buf[0] == 0xeb
	&& buf[sec_size - 2] == 0x55 && buf[sec_size - 1] == 0xaa)
    {
      return (unit_no == 0) ? 0 : -1;
    }

  /* ZIP Iomega */
  else if (memcmp ("IPAR", buf + 2, 4) == 0)
    {
      return (unit_no == 0) ? 0 : -1;
    }

  return -1;
}


/* Function */
SDR*
SDRopen (const char* filename)
{
  int unit_no, sec_size, top_blk;
  SDR* sdr;
  SDR* new_sdr;
  int* p;
  int rc;

  if ((sdr = malloc (sizeof (SDR) + 32*1024)) == NULL)
    return NULL;

  if ((rc = _dos_fatchk2 (filename, (unsigned short*) sdr->list - 1, 32*1024)) < 0)
    {
      free (sdr);
      errno = _errcnv (rc);
      return NULL;
    }
  if ((new_sdr = realloc (sdr, sizeof (SDR) + rc - 2)) == NULL)
    {
      free (sdr);
      return NULL;
    }
  sdr = new_sdr;

  {
    int drive = *((unsigned short*) sdr->list - 1);
    struct _dpbptr dpb;

    if ((rc = _dos_getdpb (drive, &dpb)) < 0
     || (rc = _dos_ioctrlrd (drive, (char*) &sdr->id, 4)) < 0)
      {
	errno = _errcnv (rc);
	free (sdr);
	return NULL;
      }
    unit_no = dpb.unit;
    sec_size = dpb.byte;
  }

  {
    struct _readcap cap;

    if ((rc = _scsi_readcap (sdr->id, &cap)) < 0)
      {
	errno = _errcnv (rc);
	free (sdr);
	return NULL;
      }

    /* ubNTCY 256<<size ̌`ɕϊ */
    switch (cap.size)
      {
      case 256:
	sdr->size = 0;
	break;
      case 512:
	sdr->size = 1;
	break;
      case 1024:
	sdr->size = 2;
	break;
      case 2048:
	sdr->size = 3;
	break;
      default:
	free (sdr);
	errno = EPERM;			/* K */
	return NULL;
      }
  }

  if ((top_blk = get_top_blk (unit_no, sec_size, sdr->id, sdr->size)) < 0)
    {
      errno = EPERM;			/* K */
      free (sdr);
      return NULL;
    }

  for (p = sdr->list; *p;)
    {
      int n;

      n = (*p * sec_size) >> 8 >> sdr->size;
      *p++ = n + top_blk;
      n = (*p * sec_size) >> 8 >> sdr->size;
      *p++ = n;
    }

  sdr->buf_rest = 0;
  sdr->blk_rest = 0;
  sdr->next = sdr->list;

  return sdr;
}


/* Function */
int
SDRclose (SDR* sdr)
{
  if (sdr == NULL)
    return -1;

  free (sdr->list);
  free (sdr);
  return 0;
}


/* Function */
int
SDRread (SDR* sdr, void* ptr, int size)
{
  int r = size;
  int blk;

  if (sdr == NULL)
    return 0;

  /* obt@ɎcĂ΍ŏɂo͂ */
  if (sdr->buf_rest)
    {
      int n = min (r, sdr->buf_rest);

      memcpy (ptr, sdr->buf + sizeof (sdr->buf) - sdr->buf_rest, n);
      sdr->buf_rest -= n;
      ptr += n;
      r -= n;

      if (r == 0)
	return size;
    }

  /* ubNPʂŏ */
  blk = r >> 8 >> sdr->size;
  while (blk)
    {
      int b;

      /* ̃`FC */
      if (sdr->blk_rest == 0)
	{
	  int* p = sdr->next;

	  if (*p == 0)
	    return 0;
	  sdr->blk_no = *p++;
	  sdr->blk_rest = *p++;
	  sdr->next = p;
	}

      /* ̃`FCǂݍރubN */
      b = min (blk, sdr->blk_rest);

      if (_scsi_readext (sdr->blk_no, b, sdr->id, sdr->size, ptr) < 0)
	return 0;
      sdr->blk_no += b;
      sdr->blk_rest -= b;
      ptr += b << 8 << sdr->size;
      blk -= b;
    }

  /* ubN */
  r &= (256 << sdr->size) - 1;
  if (r)
    {
      /* ̃`FC */
      if (sdr->blk_rest == 0)
	{
	  int* p = sdr->next;

	  if (*p == 0)
	    return 0;
	  sdr->blk_no = *p++;
	  sdr->blk_rest = *p++;
	  sdr->next = p;
	}

      /* obt@ɓǂݍŕKvȕo͂ */
      if (_scsi_readext (sdr->blk_no, 1, sdr->id, sdr->size, sdr->buf) < 0)
	return 0;
      ++sdr->blk_no;
      --sdr->blk_rest;

      memcpy (ptr, sdr->buf, r);
      sdr->buf_rest = (256 << sdr->size) - r;
    }

  return size;
}


/* EOF */
